#include <stdio.h>
#include <stdlib.h>
#include <mpi.h>

static const int root = 0;

#define N 4
#define T 2
static int m[N][N];

int main(int argc, char *argv[]) {
  int sz, me;
  int i, j, k, ib, jb, mode;

  MPI_Init(&argc, &argv);

  mode = (argc > 1 ? atoi(argv[1]) : 0);

  MPI_Comm_size(MPI_COMM_WORLD, &sz);
  MPI_Comm_rank(MPI_COMM_WORLD, &me);

  if (me == root) {
    for (j = 0; j < N; j++)
      for (i = 0; i < N; i++)
        m[j][i] = j*N + i;
  } else {
    for (j = 0; j < N; j++)
      for (i = 0; i < N; i++)
        m[j][i] = -1;
  }

  if (mode == 0) {
    /* In terms of INTs: */
    int block[T][T];
    if (me == root) {
      for (k = 0; k < sz; k++) {
        if (k != root) {
          ib = k % (N/T);
          jb = k / (N/T);

          for (j = 0; j < T; j++)
            for (i = 0; i < T; i++)
              block[j][i] = m[jb*T + j][ib*T + i];
          MPI_Send(block, T*T, MPI_INT, k, 8, MPI_COMM_WORLD);
        }
      }
    } else {
      MPI_Status s;
      MPI_Recv(block, T*T, MPI_INT, root, 8, MPI_COMM_WORLD, &s);
      
      for (j = 0; j < T; j++)
        for (i = 0; i < T; i++)
          m[j][i] = block[j][i];
    }
  } else {
    /* In terms of rows (same result): */
    MPI_Datatype block;
    
    MPI_Type_vector(T, T, N, MPI_INT, &block);
    MPI_Type_commit(&block);

    if (mode == 1) {
      if (me == root) {
        for (k = 0; k < sz; k++) {
          if (k != root) {
            ib = k % (N/T);
            jb = k / (N/T);
            for (i = 0; i < sz; i++)
              MPI_Send(&m[jb*T][ib*T], 1, block, k, 8, MPI_COMM_WORLD);
          }
        }
      } else {
        MPI_Status s;
        MPI_Recv(&m[0][0], 1, block, root, 8, MPI_COMM_WORLD, &s);
      }
    } else {
      /* No way to use scatter, because the offset from one block to
         another is not a constant. */
    }

    MPI_Type_free(&block);
  }
  
  printf("At %d:\n", me);
  for (j = 0; j < N; j++) {
    for (i = 0; i < N; i++)
      printf(" %2d", m[j][i]);
    printf("\n");
  }
  printf("\n");

  MPI_Finalize();
}
