#include <stdio.h>
#include <stdlib.h>
#include <mpi.h>

static const int root = 0;

#define M 3
#define N 4
static int m[M][N];

int main(int argc, char *argv[]) {
  int sz, me;
  int i, j, mode;

  MPI_Init(&argc, &argv);

  mode = (argc > 1 ? atoi(argv[1]) : 0);

  MPI_Comm_size(MPI_COMM_WORLD, &sz);
  MPI_Comm_rank(MPI_COMM_WORLD, &me);

  if (me == root) {
    for (j = 0; j < M; j++)
      for (i = 0; i < N; i++)
        m[j][i] = j*N + i;
  } else {
    for (j = 0; j < M; j++)
      for (i = 0; i < N; i++)
        m[j][i] = -1;
  }

  if (mode == 0) {
    /* In terms of INTs: */
    int col[M];
    if (me == root) {
      for (i = 0; i < sz; i++) {
        if (i != root) {
          for (j = 0; j < M; j++)
            col[j] = m[j][i];
          MPI_Send(col, M, MPI_INT, i, 8, MPI_COMM_WORLD);
        }
      }
    } else {
      MPI_Status s;
      MPI_Recv(col, M, MPI_INT, root, 8, MPI_COMM_WORLD, &s);
      for (j = 0; j < M; j++)
        m[j][0] = col[j];
    }
  } else {
    /* In terms of rows (same result): */
    MPI_Datatype one_col;

    MPI_Type_vector(M, 1, N, MPI_INT, &one_col);
    MPI_Type_commit(&one_col);

    if (mode == 1) {
      if (me == root) {
        for (i = 0; i < sz; i++)
          if (i != root)
            MPI_Send(&m[0][i], 1, one_col, i, 8, MPI_COMM_WORLD);
      } else {
        MPI_Status s;
        MPI_Recv(&m[0][0], 1, one_col, root, 8, MPI_COMM_WORLD, &s);
      }
    } else {
      /* Same thing using scatter: */
      MPI_Datatype col;
      int lens[2] = { 1, 1 };
      MPI_Aint offsets[2] = { 0, sizeof(int) };
      MPI_Datatype types[2] = { one_col, MPI_UB };
      
      MPI_Type_struct(2, lens, offsets, types, &col);
      MPI_Type_commit(&col);

      MPI_Scatter(m, 1, col, m, 1, col, root, MPI_COMM_WORLD);

      MPI_Type_free(&col);
    }

    MPI_Type_free(&one_col);
  }
  
  printf("At %d:\n", me);
  for (j = 0; j < M; j++) {
    for (i = 0; i < N; i++)
      printf(" %2d", m[j][i]);
    printf("\n");
  }
  printf("\n");

  MPI_Finalize();
}
