#include <stdio.h>
#include <stdlib.h>
#include <mpi.h>

static const int root = 0;

#define N 12
static int array[N];

static int sz, me;

/************************************************************/

static int *ecnt; /* number of elements (out of N) for each process */
static int *eoff; /* offset for elements of each process */

static void init_ecnt()
{
  int i;

  ecnt = (int *)malloc(sizeof(int) * sz);
  eoff = (int *)malloc(sizeof(int) * sz);
  
  for (i = 0; i < sz; i++) {
    ecnt[i] = N / sz;
    if (i == 0)
      eoff[i] = 0;
    else
      eoff[i] = eoff[i-1] + ecnt[i-1];
  }
  /* Give leftovers to last process: */
  ecnt[sz - 1] += (N - sz * (N / sz));
}

/************************************************************/

int main(int argc, char *argv[]) {
  MPI_Init(&argc, &argv);

  MPI_Comm_size(MPI_COMM_WORLD, &sz);
  MPI_Comm_rank(MPI_COMM_WORLD, &me);

  init_ecnt();

  {
    int i;

    if (me == root) {
      for (i = 0; i < N; i++)
        array[i] = i * i;
    } else {
      for (i = 0; i < N; i++)
        array[i] = -1;
    }

    MPI_Scatterv(array, ecnt, eoff, MPI_INT, array, ecnt[me], MPI_INT, 
                 root, MPI_COMM_WORLD);

    printf("At %d:", me);
    for (i = 0; i < N; i++)
      printf(" %d", array[i]);
    printf("\n");

    for (i = 0; i < ecnt[me]; i++)
      array[i] += me + 1;
    
    if (argc == 1)
      MPI_Gatherv(array, ecnt[me], MPI_INT, array, ecnt, eoff, MPI_INT, 
                  root, MPI_COMM_WORLD);
    else
      MPI_Allgatherv(array, ecnt[me], MPI_INT, array, ecnt, eoff, MPI_INT,
                     MPI_COMM_WORLD);
    
    printf("Final at %d:", me);
    for (i = 0; i < N; i++)
      printf(" %d", array[i]);
    printf("\n");
  }

  MPI_Finalize();
}
