#include <stdio.h>
#include <stdlib.h>
#include <mpi.h>

static const int root = 0;

#define N 3
#define M 4
static int m[M][N];

int main(int argc, char *argv[]) {
  int sz, me;
  int i, j, mode;

  MPI_Init(&argc, &argv);

  mode = (argc > 1 ? atoi(argv[1]) : 0);

  MPI_Comm_size(MPI_COMM_WORLD, &sz);
  MPI_Comm_rank(MPI_COMM_WORLD, &me);

  if (me == root) {
    for (j = 0; j < M; j++)
      for (i = 0; i < N; i++)
        m[j][i] = j*N + i;
  } else {
    for (j = 0; j < M; j++)
      for (i = 0; i < N; i++)
        m[j][i] = -1;
  }

  if (mode == 0) {
    /* In terms of INTs: */
    if (me == root) {
      for (j = 0; j < sz; j++)
        if (j != root)
          MPI_Send(&m[j][0], N, MPI_INT, j, 8, MPI_COMM_WORLD);
    } else {
      MPI_Status s;
      MPI_Recv(&m[0][0], N, MPI_INT, root, 8, MPI_COMM_WORLD, &s);
    }
  } else {
    /* In terms of rows (same result): */
    MPI_Datatype row;

    MPI_Type_contiguous(N, MPI_INT, &row);
    MPI_Type_commit(&row);

    if (mode == 1) {
      if (me == root) {
        for (j = 0; j < sz; j++)
          if (j != root)
            MPI_Send(&m[j][0], 1, row, j, 8, MPI_COMM_WORLD);
      } else {
        MPI_Status s;
        MPI_Recv(&m[0][0], 1, row, root, 8, MPI_COMM_WORLD, &s);
      }
    } else {
      /* Same thing using scatter: */
      MPI_Scatter(m, 1, row, m, 1, row, root, MPI_COMM_WORLD);
    }

    MPI_Type_free(&row);
  }
  
  printf("At %d:\n", me);
  for (j = 0; j < M; j++) {
    for (i = 0; i < N; i++)
      printf(" %2d", m[j][i]);
    printf("\n");
  }
  printf("\n");

  MPI_Finalize();
}
